import os.path
import multiprocessing
import shutil

from sys import platform
from glob import glob
from anasymod.files import get_full_path, mkdir_p
from anasymod.util import vivado_search_key
from os import environ as env
from anasymod.enums import BoardNames
from anasymod.plugins import *
from anasymod.fpga_boards.boards import *
from anasymod.base_config import BaseConfig

class EmuConfig:
    def __init__(self, root, cfg_file, build_root=None):

        self.root = root

        # define and create build root
        self.build_root = build_root if build_root is not None else os.path.join(root, 'build')
        if not os.path.exists(self.build_root):
            mkdir_p(self.build_root)

        self.cfg_file = cfg_file

        # Initialize config  dict
        self.cfg = Config(cfg_file=self.cfg_file)

        # Update config options by reading from config file
        self.cfg.update_config()

        # Try to initialize Inicio config_dict
        # TODO: add more reasonable defaults
        try:
            from inicio import config_dict
            self.cfg_dict = config_dict()
        except:
            self.cfg_dict = {
                'TOOLS_xilinx': '',
                'INICIO_TOOLS': '',
                'TOOLS_iverilog': '',
                'TOOLS_gtkwave': ''
            }

        # FPGA board configuration
        self.board = self._fetch_board()

        # Vivado configuration
        self.vivado_config = VivadoConfig(parent=self)

        # GtkWave configuration
        self.gtkwave_config = GtkWaveConfig(parent=self)

        # SimVision configuration
        self.simvision_config = SimVisionConfig(parent=self)

        # Icarus configuration
        self.icarus_config = IcarusConfig(parent=self)

        # Xcelium configuration
        self.xcelium_config = XceliumConfig(parent=self)

    @property
    def ila_depth(self):
        return 4096

    def _fetch_board(self):
        """
        Fetch FPGA board info.
        """

        if self.cfg.board_name == BoardNames.PYNQ_Z1:
            return PYNQ_Z1()
        elif self.cfg.board_name == BoardNames.ARTY_A7:
            return ARTY_A7()
        elif self.cfg.board_name == BoardNames.VC707:
            return VC707()
        elif self.cfg.board_name == BoardNames.ULTRA96:
            return ULTRA96()
        elif self.cfg.board_name == BoardNames.TE0720:
            return TE0720()
        elif self.cfg.board_name == BoardNames.ZC702:
            return ZC702()
        else:
            raise Exception(f'The requested board {self.cfg.board_name} could not be found.')

class VivadoConfig():
    def __init__(self, parent: EmuConfig, vivado=None):
        # save reference to parent config
        self.parent = parent

        # set project name
        self.project_name = 'project'
        # intermediate variables for generic Xilinx path
        if platform in {'win32', 'cygwin'}:
            xilinx_version_path = parent.cfg_dict['TOOLS_xilinx']
            xilinx_version = "20" + ".".join(xilinx_version_path.split(".")[0:2]).split("-")[1]
        # set path to vivado binary
        self.hints = [lambda: os.path.join(env['VIVADO_INSTALL_PATH'], 'bin'),
                      lambda: os.path.join(parent.cfg_dict['INICIO_TOOLS'], xilinx_version_path, "Vivado", xilinx_version, "bin" ),]

        if platform in {'linux', 'linux2'}:
            sorted_dirs = sorted(glob('/tools/Xilinx/Vivado/*.*'), key=vivado_search_key)
            self.hints.extend(lambda: os.path.join(dir_, 'bin') for dir_ in sorted_dirs)

        self._vivado = vivado
        # set various project options
        self.num_cores = multiprocessing.cpu_count()
        self.vio_name = 'vio_0'
        self.vio_inst_name = self.vio_name + '_i'
        self.ila_inst_name = 'u_ila_0'
        self.ila_reset = 'reset_probe'
        self.vio_reset = 'vio_i/rst'

    @property
    def vivado(self):
        if self._vivado is None:
            self._vivado = find_tool(name='vivado', hints=self.hints)
        return self._vivado

class XceliumConfig():
    def __init__(self, parent: EmuConfig, xrun=None):
        # save reference to parent config
        self.parent = parent

        # set path to xrun binary
        self.hints = [lambda: os.path.join(env['XCELIUM_INSTALL_PATH'], 'bin')]
        self._xrun = xrun

        # name of TCL file
        self.tcl_input = 'input.tcl'

    @property
    def xrun(self):
        if self._xrun is None:
            try:
                self._xrun = find_tool(name='ifxxcelium', hints=self.hints)
                #self._xrun += " execute"
            except:
                self._xrun = find_tool(name='xrun', hints=self.hints)
        return self._xrun

    @property
    def tcl_input_path(self):
        return os.path.join(self.parent.build_root, self.tcl_input)

class IcarusConfig():
    def __init__(self, parent: EmuConfig, iverilog=None, vvp=None):
        # save reference to parent config
        self.parent = parent

        # set path to iverilog and vvp binaries
        self.hints = [lambda: os.path.join(env['ICARUS_INSTALL_PATH'], 'bin'),
                      lambda: os.path.join(parent.cfg_dict['INICIO_TOOLS'], parent.cfg_dict['TOOLS_iverilog'], 'bin')]
        self._iverilog = iverilog
        self._vvp = vvp

        # name of output file
        self.output_file_name = 'a.out'

    @property
    def iverilog(self):
        if self._iverilog is None:
            self._iverilog = find_tool(name='iverilog', hints=self.hints)
        return self._iverilog

    @property
    def vvp(self):
        if self._vvp is None:
            self._vvp = find_tool(name='vvp', hints=self.hints)
        return self._vvp

    @property
    def output_file_path(self):
        return os.path.join(self.parent.build_root, self.output_file_name)

class GtkWaveConfig():
    def __init__(self, parent: EmuConfig, gtkwave=None):
        # save reference to parent config
        self.parent = parent

        # find gtkwave binary
        self.hints = [lambda: os.path.join(env['GTKWAVE_INSTALL_PATH'], 'bin'),
                      lambda: os.path.join(parent.cfg_dict['INICIO_TOOLS'], parent.cfg_dict['TOOLS_gtkwave'], 'bin')]
        self._gtkwave = gtkwave
        self.gtkw_config = None

    @property
    def gtkwave(self):
        if self._gtkwave is None:
            self._gtkwave = find_tool(name='gtkwave', hints=self.hints)
        return self._gtkwave

class SimVisionConfig():
    def __init__(self, parent: EmuConfig, simvision=None):
        # save reference to parent config
        self.parent = parent

        # find simvision binary
        self.hints = [lambda: os.path.join(env['SIMVISION_INSTALL_PATH'], 'bin')]
        self._simvision = simvision
        self.svcf_config = None

    @property
    def simvision(self):
        if self._simvision is None:
            self._simvision = find_tool(name='simvision', hints=self.hints)
        return self._simvision

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """
    def __init__(self, cfg_file):
        super().__init__(cfg_file=cfg_file, section=ConfigSections.PROJECT)
        self.dec_bits = 24
        self.board_name = BoardNames.PYNQ_Z1
        #self.board_name = BoardNames.VC707
        self.emu_clk_freq = 25e6
        self.preprocess_only = False
        self.jtag_freq = 15e6
        self.fpga_sim_crtl = FPGASimCtrl.VIVADO_VIO
        self.plugins = []
        self.plugins.append('msdsl')
        self.dt = 0.1e-6
        self.dt_exponent = -46
        self.dt_width = 27
        self.time_width = 64

def find_tool(name, hints=None, sys_path_hint=True):
    # set defaults
    if hints is None:
        hints = []



    # add system path as the last "hint" if desired (default behavior)
    if sys_path_hint:
        hints.append(lambda: None)

    # first check the system path for the tool
    tool_path = shutil.which(name)

    # if the tool isn't found in the system path, then try out the hints in order
    if tool_path is None:
        for hint in hints:
            try:
                if callable(hint):
                    tool_path = shutil.which(name, path=hint())
                else:
                    tool_path = shutil.which(name, path=hint)
            except:
                continue

            if tool_path is not None:
                break

    if tool_path is not None:
        return get_full_path(tool_path)
    else:
        raise KeyError(f'Tool:{name} could not be found')





import os

from anasymod.defines import Define
from anasymod.util import back2fwd
from anasymod.config import EmuConfig
from anasymod.base_config import BaseConfig
from anasymod.enums import ConfigSections, FPGASimCtrl
from anasymod.structures.structure_config import StructureConfig
from anasymod.structures.module_top import ModuleTop
from anasymod.structures.module_clk_manager import ModuleClkManager
from typing import Union
from anasymod.sources import VerilogSource
from anasymod.sim_ctrl.uart_ctrlinfra import UARTControlInfrastructure
from anasymod.sim_ctrl.vio_ctrlinfra import VIOControlInfrastructure
from anasymod.structures.module_traceport import ModuleTracePort

class Target():
    """
    This class inherits all source and define objects necessary in order to run actions for a specific target.

    Attributes:
        _name               Target name
        content    Dict of Lists of source and define objects associated with target
    """
    def __init__(self, prj_cfg: EmuConfig, plugins: list, name):
        self.prj_cfg = prj_cfg
        self.plugins = plugins

        # Initialize structure configuration
        self.str_cfg = StructureConfig(prj_cfg=self.prj_cfg)

        # Instantiate Simulation ControlInfrastructure Interface
        self.ctrl = None
        """ :type : ControlInfrastructure"""

        self._name = name

        # Initialize Probe Objs; there can be a probe obj for each supported simulator and synthesizer
        self.probes = {}
        """ :type : dict(Probe)"""

        # Initialize content dict  to store design sources and defines
        self.content = Content()

        # Initialize target_config
        self.cfg = Config(cfg_file=self.prj_cfg.cfg_file, prj_cfg=self.prj_cfg, name=self._name)

    def set_tstop(self):
        """
        Add define statement to specify tstop
        """
        self.content.defines.append(Define(name='TSTOP_MSDSL', value=self.cfg.tstop))

    def assign_fileset(self, fileset: dict):
        """
        Add a fileset to the target by adding defines and sources to the according attributes

        :param fileset: fileset that shall be added to target
        :type fileset: dict
        """

        for k in fileset.keys():
            setattr(self.content, k, getattr(self.content, k) + fileset[k])

    def update_structure_config(self):
        self.str_cfg.cfg.update_config(subsection=self._name)

    def gen_structure(self):
        """
        Generate toplevel, IPCore wrappers, debug infrastructure and clk manager.
        """

        # Generate toplevel and add to target sources
        toplevel_path = os.path.join(self.prj_cfg.build_root, 'gen_top.sv')
        with (open(toplevel_path, 'w')) as top_file:
            top_file.write(ModuleTop(target=self).render())

        self.content.verilog_sources += [VerilogSource(files=toplevel_path)]

        # Build control structure and add all sources to project
        if self.ctrl is not None:
            self.ctrl.gen_ctrlwrapper(str_cfg=self.str_cfg, content=self.content)

        # Generate clk management wrapper and add to target sources
        clkmanagerwrapper_path = os.path.join(self.prj_cfg.build_root, 'gen_clkmanager_wrap.sv')
        with (open(clkmanagerwrapper_path, 'w')) as clkm_file:
            clkm_file.write(ModuleClkManager(scfg=self.str_cfg).render())

        self.content.verilog_sources += [VerilogSource(files=clkmanagerwrapper_path)]

        # Generate traceport wrapper and add to target sources
        trapwrapper_path = os.path.join(self.prj_cfg.build_root, 'gen_traceport_wrap.sv')
        with (open(trapwrapper_path, 'w')) as trap_file:
            trap_file.write(ModuleTracePort(scfg=self.str_cfg).render())

        self.content.verilog_sources += [VerilogSource(files=trapwrapper_path)]

    def setup_ctrl_ifc(self):
        """
        Setup the control interface according to what was provided in the project configuration. default is VIVADO_VIO
        mode, which does not possess a direct control interface via anasymod.

        :rtype: ControlInfrastructure
        """
        #ToDo: This should only be executed for FPGA simulation!

        if self.cfg.fpga_sim_ctrl == FPGASimCtrl.VIVADO_VIO:
            print("No direct control interface from anasymod selected, Vivado VIO interface enabled.")
            self.ctrl = VIOControlInfrastructure(prj_cfg=self.prj_cfg)
        elif self.cfg.fpga_sim_ctrl == FPGASimCtrl.UART_ZYNQ:
            print("Direct anasymod FPGA simulation control via UART enabled.")
            self.ctrl = UARTControlInfrastructure(prj_cfg=self.prj_cfg)
        else:
            raise Exception("ERROR: No FPGA simulation control was selected, shutting down.")

    @property
    def project_root(self):
        return os.path.join(self.prj_cfg.build_root, self.prj_cfg.vivado_config.project_name)

class SimulationTarget(Target):
    def __init__(self, prj_cfg: EmuConfig, plugins: list, name=r"sim"):
        super().__init__(prj_cfg=prj_cfg, plugins=plugins, name=name)

    def setup_vcd(self):
        self.content.defines.append(Define(name='VCD_FILE_MSDSL', value=back2fwd(self.cfg.vcd_path)))

class FPGATarget(Target):
    """

    Attributes:
        _ip_cores        List of ip_core objects associated with target, those will generated during the build process
    """
    def __init__(self, prj_cfg: EmuConfig, plugins: list, name=r"fpga"):
        # call the super constructor
        super().__init__(prj_cfg=prj_cfg, plugins=plugins, name=name)

        # use a different default TSTOP value, which should provide about 0.1 ps timing resolution and plenty of
        # emulation time for most purposes.
        self._ip_cores = []

    def gen_structure(self):
        """
        Generate toplevel, IPCore wrappers, debug/ctrl infrastructure for FPGA and clk manager.
        """
        super().gen_structure()
        self.ctrl.gen_ctrl_infrastructure(str_cfg=self.str_cfg, content=self.content)

    @property
    def probe_cfg_path(self):
        return os.path.join(self.project_root, 'probe_config.txt')

    @property
    def bitfile_path(self):
        return os.path.join(self.project_root, f'{self.prj_cfg.vivado_config.project_name}.runs', 'impl_1',
                            f"{self.cfg.top_module}.bit")

    @property
    def ltxfile_path(self):
        return os.path.join(self.project_root, f'{self.prj_cfg.vivado_config.project_name}.runs', 'impl_1',
                            f"{self.cfg.top_module}.ltx")

    @property
    def ip_dir(self):
        return os.path.join(self.project_root, f'{self.prj_cfg.vivado_config.project_name}.srcs', 'sources_1', 'ip')

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """
    def __init__(self, cfg_file, prj_cfg, name):
        super().__init__(cfg_file=cfg_file, section=ConfigSections.TARGET)
        self.tstop = 10.0
        self.emu_clk_freq =25e6
        self.top_module = 'top'
        self.custom_top = False
        self.vcd_name = f"{self.top_module}_{name}.vcd"
        self.vcd_path = os.path.join(prj_cfg.build_root, r"vcd", self.vcd_name)
        # TODO: move these paths to toolchain specific config, which shall be instantiated in the target class
        self.csv_name = f"{self.top_module}_{name}.csv"
        self.csv_path = os.path.join(prj_cfg.build_root, r"csv", self.csv_name)
        self.fpga_sim_ctrl = FPGASimCtrl.VIVADO_VIO

class Content():
    """
    Container storing all Project specific source files, such as RTL, eSW code, blockdiagrams, IP core configs,
    constraint files, ... .
    """

    def __init__(self):
        self.verilog_sources = []
        """:type : List[VerilogSource]"""
        self.verilog_headers = []
        """:type : List[VerilogHeader]"""
        self.vhdl_sources = []
        """:type : List[VHDLSource]"""
        self.defines = []
        """:type : List[Define]"""
        self.xci_files = []
        """:type : List[XCIFile]"""
        self.xdc_files = []
        """:type : List[XDCFile]"""
        self.mem_files = []
        """:type : List[MEMFile]"""
        self.bd_files = []
        """:type : List[BDFile]"""



class ProbeConfig:
    def __init__(self, probe_cfg_path):
        # Read Probe Config file
        with (open(probe_cfg_path, "r")) as cf:
            config = cf.readlines()

        prefixes = [r"ANALOG:", r"ANALOG_EXPONENT:", r"ANALOG_WIDTH:", r"TIME:", r"TIME_EXPONENT:", r"TIME_WIDTH:", r"RESET:", r"SB:", r"MB:", r"MB_WIDTH:"]
        # clean and remove prefixes
        signals = []
        for signal, prefix in zip(config, prefixes):
            signals.append(signal.strip(prefix).strip().split())

        self.analog_signals = []
        for name, width, exponent in zip(signals[0], signals[2], signals[1]):
            self.analog_signals.append((name, width, exponent))

        self.time_signal = []
        for name,width, exponent in zip(signals[3], signals[5], signals[4]):
            self.time_signal.append((name, width, exponent))

        self.reset_signal = []
        for name,width, exponent in zip(signals[6], [r"1"], [None]):
            self.reset_signal.append((name, width, exponent))

        self.digital_signals = []
        for name in signals[7]:
            self.digital_signals.append((name, r"1", None))

        for name, width in zip(signals[8], signals[9]):
            self.digital_signals.append((name, width, None))




from anasymod.enums import ConfigSections

class BaseConfig():
    """
    Base class for a config container class. Includes all the base functions that shall be applicable on a config class,
    e.g. update from the prj_config file, read from prj_config file ... .
    """

    def __init__(self, cfg_file: dict, section):
        self.cfg_file = cfg_file
        if section in ConfigSections.__dict__.keys():
            self.section = section
        else:
            raise KeyError(f"provided section key:{section} is not supported")

    def update_config(self, subsection=None):
        """
        Update config by entries from config file; datatype is a dict.
        Using subsection argument, it is possible to read an object specific config section, this
        is necessary if multiple objects of one class are in use, e.g. for multiple targets.

        :param subsection: subsection to use from config file
        """
        if self.cfg_file is not None:
            if self.section in self.cfg_file:
                if subsection is not None:
                    config_section = self.cfg_file[self.section].get(subsection)
                else:
                    config_section = self.cfg_file.get(self.section)

                # update attributes from config class
                if config_section is not None:
                    for k, v in config_section.items():
                        if hasattr(self, k):
                            setattr(self, k, v)
                        else:
                            print(f"Warning: Processing config section:{config_section}; provided key: {k} does not exist in config")



import os

from anasymod.sources import Sources, VerilogHeader, VerilogSource, VHDLSource
from anasymod.defines import Define
from anasymod.enums import ConfigSections
from anasymod.base_config import BaseConfig

class Plugin():
    def __init__(self, cfg_file, prj_root, build_root, name):
        self._cfg_file = cfg_file
        self._prj_root = prj_root
        self._build_root = build_root
        self._srccfg_path = os.path.join(self._prj_root, r"source.config")
        self._name = name
        self._defines = []
        self._verilog_sources = []
        """:type : List[VerilogSource]"""

        self._verilog_headers = []
        """:type : List[VerilogHeader]"""

        self._vhdl_sources = []
        """:type : List[VHDLSource]"""

        # List of includes that need to be added to source files generated via ANASYMOD
        self.cfg = Config(cfg_file=self._cfg_file)

        self.include_statements = []

    def _dump_defines(self):
        return self._defines

    def _dump_verilog_sources(self):
        return self._verilog_sources

    def _dump_verilog_headers(self):
        return self._verilog_headers

    def _dump_vhdl_sources(self):
        return self._vhdl_sources

    def _setup_sources(self):
        """
        Add Source objects that are specific to MSDSL
        """
        raise NotImplementedError()

    def _setup_defines(self):
        """
        Add Define objects that are specific to MSDSL
        """
        raise NotImplementedError()

    def add_source(self, source: Sources):
        if isinstance(source, VerilogSource):
            self._verilog_sources.append(source)
        if isinstance(source, VerilogHeader):
            self._verilog_headers.append(source)
        if isinstance(source, VHDLSource):
            self._vhdl_sources.append(source)

    def add_define(self, define: Define):
        self._defines.append(define)

    def _parse_args(self):
        """
        Read command line arguments. This supports convenient usage from command shell e.g.:
        python analysis.py -i filter --models --sim --view
        """
        pass

    #def _update_config(self, cfg, config_section):
    #    return self.cfg.update_config(cfg=cfg, config_section=config_section)

    #def _read_config(self, cfg_file, section, subsection=None):
    #    return read_config(cfg_file=cfg_file, section=section, subsection=subsection)

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """

    def __init__(self, cfg_file):
        super().__init__(cfg_file=cfg_file, section=ConfigSections.PLUGIN)


import numpy as np

try:
    from vcd import VCDWriter
except:
    print('ERROR: Could not load pyvcd package!')

import datetime

from anasymod.probe_config import ProbeConfig
from anasymod.targets import FPGATarget

class ConvertWaveform():
    def __init__(self, target: FPGATarget):

        # read CSV file
        with open(target.cfg.csv_path, 'r') as f:
            first_line = f.readline()

        # split up the first line into comma-delimited names
        signals = first_line.split(',')

        # account for whitespace
        signals = [signal.strip() for signal in signals]

        # strip off the signal indices and add to a lookup table
        signal_lookup = {}
        for k, signal in enumerate(signals):
            if '[' in signal:
                signal = signal[:signal.index('[')]
            signal_lookup[signal] = k

        # print keys
        print(f'CSV column names: {[key for key in signal_lookup.keys()]}')

        # define method for getting unscaled data
        def get_csv_col(name):
            return np.genfromtxt(target.cfg.csv_path, delimiter=',', usecols=signal_lookup[name], skip_header=1)

        # read probe signal information to find out what signals are analog, digital, reset, time, etc.
        signals = ProbeConfig(probe_cfg_path=target.probe_cfg_path)

        # store data from FPGA in a dictionary
        probe_data = {}
        real_signals = set()
        reg_widths = {}

        # Add analog signals to probe_data
        for name, _, exponent in signals.analog_signals + signals.time_signal:
            if name not in signal_lookup:
                continue

            # add to set of probes with "real" data type
            real_signals.add(name)

            # get unscaled data and apply scaling factor
            probe_data[name] = (2**int(exponent)) * get_csv_col(name)

            # convert data to native Python float type (rather than numpy float)
            # this is required for PyVCD
            probe_data[name] = [float(x) for x in probe_data[name]]

        # Add digital signals to probe_data
        for name, width, _ in signals.digital_signals + signals.reset_signal:
            if name not in signal_lookup:
                continue

            # define width for this probe
            reg_widths[name] = int(width)

            # get unscaled data
            probe_data[name] = get_csv_col(name)

            # convert data to native Python float type (rather than numpy int)
            # this is required for PyVCD
            probe_data[name] = [int(x) for x in probe_data[name]]

        # Extract the time signal from the probe data since it is needed to produce a VCD file
        time_signal_name, _, _ = signals.time_signal[0]
        time_signal = probe_data[time_signal_name]

        # Write data to VCD file
        with open(target.cfg.vcd_path, 'w') as vcd:
            with VCDWriter(vcd, timescale='1 ns', date=str(datetime.datetime.today())) as writer:
                # register all of the signals that will be written to VCD
                reg = {}
                for signal_full_name, scaled_data in probe_data.items():
                    # determine signal scope and name
                    signal_split = signal_full_name.split('/')
                    vcd_scope = '.'.join(signal_split[:-1])
                    vcd_name = signal_split[-1]

                    # determine signal type and size
                    if signal_full_name in real_signals:
                        vcd_var_type = 'real'
                        vcd_size = None
                    elif signal_full_name in reg_widths:
                        vcd_var_type = 'reg'
                        vcd_size = reg_widths[signal_full_name]
                    else:
                        raise Exception('Unknown signal type.')

                    # register the signal
                    reg[signal_full_name] = writer.register_var(scope=vcd_scope, name=vcd_name, var_type=vcd_var_type,
                                                                size=vcd_size)

                # iterate over all timesteps
                for k, timestamp in enumerate(time_signal):
                    # break if timestamp is less than zero since it means that wrapping has occurred
                    if timestamp < 0:
                        break

                    # iterate over all signals and log their change at this timestamp
                    for signal_full_name, scaled_data in probe_data.items():
                        writer.change(reg[signal_full_name], round(1e9 * timestamp), scaled_data[k])


class ConfigSections:
    """
    Container including enums for sections that can be used in config file.
    """
    TARGET = "TARGET"
    FPGA_BOARD = "FPGA_BOARD"
    TOOL = "TOOL"
    PLUGIN = "PLUGIN"
    PROJECT = "PROJECT"
    STRUCTURE = "STRUCTURE"
    FPGASIM = "FPGASIM"

class BoardNames:
    """
    Container including enums for all supported boards.

    Currently supported boards are:
    PYNQ_Z1
    ARTY_A7
    VC707
    ZC702
    ULTRA96
    TE0720
    """
    PYNQ_Z1 = 'PYNQ_Z1'
    ARTY_A7 = 'ARTY_A7'
    VC707 = 'VC707'
    ZC702 = 'ZC702'
    ULTRA96 = 'ULTRA96'
    TE0720 = 'TE0720'

class PortDir:
    """
    Container including enums for all supported port directions.
    """
    IN = "input"
    OUT = "output"
    INOUT = "inout"

class CtrlOps:
    """
    Container including enums for all supported control operations for controlling the FPGA.
    """
    WRITE_PARAMETER = 0
    READ_PARAMETER = 1

class FPGASimCtrl:
    """
    Container including enums for all supported FPGA control interfaces.
    """
    UART_ZYNQ = 'uart_zynq'
    VIVADO_VIO = 'vivado_vio'



import numpy as np
import os
import csv
import time


from anasymod.utils.VCD_parser import ParseVCD
from typing import Union

from anasymod.targets import SimulationTarget, FPGATarget

class Probe():
    """
    Base class for simulation data access API
    """
    def __init__(self, target: Union[SimulationTarget, FPGATarget]):
        """
        Constructor
        """
        self.handle = None
        self.target = target

        #self.__ref_analysis = weakref.proxy(ref_analysis)
        #if False:
        #    from ...exported.analysis import Analysis
        #    self.__ref_analysis = Analysis()

        #self.simOpts = self.__ref_analysis.getSimOptions()
        #""":type: _Options_Base"""

        #if self.simOpts.idnum is not None:
        #    self.ident_num = self.simOpts.idnum
        #else:
        #    self.ident_num = 0

        self._data_valid = False

        #self.simulatorSimetrix = self.__ref_analysis.simulatorSimetrix
        #""":type: Simetrix"""

    def __del__(self):
        self.discardloadedsimdatafiles()

    def _probe(self, name, emu_time, cache=True):
        """
        Access probed waveform trace(s)

        :param name: name of probe whose waveform is require.
            (None = timebase reference trace)
        :type name: string
        :param cache: Cache probe data so subsequent calls for same data don't create
            more copies or trigger a SIMetrix data group load
        :type cache: bool

        :return: probed data for specified probe.
        :rtype: numpy.array
        """

        raise NotImplementedError()

    def _probes(self):
        """
        Get list of names probe waveforms for data group/ run
        :rtype: list[str]
        """

        raise NotImplementedError()

    def init_rundata(self):
        raise NotImplementedError()

    def discardloadedsimdatafile(self, run_num, check_loaded=True):
        """
        Discard probe data group from SIMetrix

        Normally a probed Data group accessed via probe is kept loaded in SIMetrix
        memory to avoid having to (expensively) reload it if further probed data vectors
        are accessed.  Calling this forces it to be unloaded from SIMetrix free-ing up
        memory.  Of course a follow-in accesses will be slow/expensive as data group
        needs to be reloaded.

        :param run_num: Run number in sweep, None for last run
        :type run_num: int
        :param check_loaded: Check group is actually loaded in SIMetrix, do nothing if it is not
        :type check_loaded: bool
        """
        raise NotImplementedError()

    def discardloadedsimdatafiles(self):
        """
        discard all   data groups currently loaded into SIMetrix
        """
        # print("DISCARD...")
        raise NotImplementedError()

    def refreshLoadedDataGroups(self):
        """
        Update local view of probe data groups currently loaded in SIMetrix

        Refresh  table of data groups loaded into SIMetrix from
        SIMetrix internal list.  Used to re-sync if
        activity in SIMetrox may have caused data groups loaded to change.
        """
        raise NotImplementedError()

    def _compress(self, wave=np.ndarray):
        """Compresses redundant data from 2d numpy array"""
        raise NotImplementedError()

    def parse_emu_time(self, data, emu_time):
        """

        :param data:
        :param emu_time:
        :return:
        """
        raise NotImplementedError()

class ProbeCSV(Probe):
    """
    API for CSV remote data access
    """

    def __init__(self, target: Union[SimulationTarget, FPGATarget]):
        """
        Constructor
        """
        super().__init__(self, target=target)

        # List of dictionaries containing probename-probevalue pairs for each simulation run (one dict per sweep point)
        self.probe_caches = []
        """:type : list[dict[str, list[]]]"""

    def __del__(self):
        """
        Destructor

        Remote Data interface is automatically closed when Api is destructed
        """

    def init_rundata(self):
        self.probe_caches = [None] * 1 # As of now no sweeps are planned, later the value shall be set to the number of runs
        self._data_valid = True

    def path_for_sim_result_file(self):
        return os.path.join(self.target.cfg['csv_path'])

    def _probe(self, name, emu_time, cache=True):
        """
        Access csv logfile data for specified run number simulation parameter

        :param name: Column name in csv log, omit/None for all
        :type name: str

        :return: dict for `run_num`only or specified,
            list  for `name`  only specified,
            numpy.array for specified `run_num` and `name`,
            list of dict for run_num='all' specified
        :rtype: numpy.array | list[numpy.array] | dict[ string : numpy.array]
        """

        if not self._data_valid:
            raise LookupError("No data available (no succesful simulation run / dataset reload)")

        run_num = 0

        if run_num == 'all':
            for r in range(len(self.probe_caches)):
                if self.probe_caches[r] is None:
                    self.probe_caches[r] = self.fetch_simdata(self.path_for_sim_result_file())
        else:
            if self.probe_caches[run_num] is None:
                self.probe_caches[run_num] = self.fetch_simdata(self.path_for_sim_result_file())

        if run_num == 'all' and name is None:
            return self.probe_caches

        if run_num != 'all' and run_num >= len(self.probe_caches):
            raise ValueError("Run number must be in range [0 .. %d]" % len(self.probe_caches))

        if name is not None and run_num != 'all' and name not in self.probe_caches[run_num]:
            print("No such  name in simulation log: ", name)
            print("Available names: ", self.probe_caches[run_num].keys())
            raise LookupError("Bad probe name " + name)

        if name is None:
            return self.probe_caches[run_num]

        if run_num == 'all':
            if name not in self.probe_caches[0]:
                print("No such  name in simulation log: ", name)
                print("Available names: ", self.probe_caches[0].keys())
                raise ValueError("Bad probe name " + name)
            res = [log[name] for log in self.probe_caches]
            return res

        if name not in self.probe_caches[run_num]:
            print("No such  name in simulation log: ", name)
            print("Available names: ", self.probe_caches[run_num].keys())
            raise ValueError("Bad probe name " + name)

        return self.probe_caches[run_num][name]

    def fetch_simdata(self, csv_logfile):
        """
        Load simdata file as dictionary

        :return: Dict of numpy arrays (keys are column titles, values are column values)
        :rtype: dict[numpy.array]
        """

        with open(csv_logfile, "r") as csvfile:
            csvstrm = csv.DictReader(csvfile, delimiter=';', quotechar='\"')

            rowdict = next(csvstrm)
            csvdict = {name: [float(value)] for name, value in rowdict.items()}

            for rowdict in csvstrm:
                for name, value in rowdict.items():
                    csvdict[name].append(float(value))
        return {name: np.array(value) for name, value in csvdict.items()}

    def _probes(self):
        """
        Get list of names probe waveforms for specified run

        :rtype:list[str]
        """

        if not self._data_valid:
            raise LookupError("No data available (no succesful simulation run / dataset open)")

        run_num = 0

        if self.probe_caches[run_num] is None:
            self.probe_caches[run_num] = self.fetch_simdata(self.path_for_sim_result_file())

        return self.probe_caches[run_num].keys()

class ProbeVCD(Probe):
    def __init__(self, target: Union[SimulationTarget, FPGATarget]):
        """
        Constructor for VCD reader.
        """
        super().__init__(target=target)

        # List of dictionaries containing probename-probevalue pairs for each simulation run (one dict per sweep point)
        self.probe_caches = []
        """:type : list[dict[str, list[]]]"""

        # VCD file handle
        self.vcd_handle = dict()
        """:type: dict[str,vcd.VCDparser]"""

        self.init_rundata()

    def __del__(self):
        """
        Destructor

        Remote Data interface is automatically closed when Api is destructed
        """
        self.probe_caches = []
        pass

    def init_rundata(self):
        self.probe_caches = [{} for _ in range(10)] # this needs a fix later, shall be depenent on number of signals that were stored during simulation
        self._data_valid = True

    def parse_emu_time(self, data, emu_time):
        """
        Parse Emu_time end returns new vector with emu_time instead of cycle count
        :param data:
        :return:
        """
        emutime_data = data.copy()
        emutime_data.setflags(write=True)

        # emu_time vector index
        t = 0
        for i in range(int(len(emutime_data[0]))):
            while emutime_data[0][i] > emu_time[0][t]:
                t += 1
                # break while loop, if it reaches the end of the array of emu_time
                if t >= len(emu_time[0]):
                    break
            # break for loop because whole emu_time vector was parsed
            if t >= len(emu_time[0]):
                break
            # if there is a data point with no coresponding emu_time cyclecount, take the last common cycle count value
            if emutime_data[0][i] < emu_time[0][t]:
                emutime_data[0][i] = emu_time[1][t]
            # if cycle count of data and emu_time vector is the same, take the time value
            else:
                emutime_data[0][i] = emu_time[1][t]

        emutime_data.setflags(write=False)
        return emutime_data

    def _probe(self, name, emu_time, cache=True):
        """
        Access VCD data for specified run number simulation parameter
        :param name: Column name in csv log, omit/None for all
        :type name: str
        :param emu_time: Use emu_time as time basis or cycle_count
        :type emu_time: bool
        :param cache:
        :return:
        """
        """
                Access vcd data for specified run number simulation parameter

                :param name: Column name in csv log, omit/None for all
                :type name: str

                :return: dict for `run_num`only or specified,
                    list  for `name`  only specified,
                    numpy.array for specified `run_num` and `name`,
                    list of dict for run_num='all' specified
                :rtype: numpy.array | list[numpy.array] | dict[ string : numpy.array]

                """
        run_num = 0
        vcd_handle = self.setup_data_access()
        try:
            run_cache = self.probe_caches[run_num]
        except IndexError:  # If PyVerify single run: run_num may be bigger than len(self.probe_caches)
            run_cache = []
            cache = False

        # check if empty run_cache and parse all signals and store in run_cache
        if len(run_cache) == 0 and cache:
            run_cache = self.fetch_simdata(vcd_handle, update_data=False)
            self.probe_caches[run_num] = run_cache

        #check complete name of emu_time_probe
        matching = [s for s in self._probes() if 'emu_time_probe' in s]
        if len(matching) == 1:
            emu_time_probe = matching[0]

        if emu_time_probe not in run_cache:
            data = self.fetch_simdata(vcd_handle, name=emu_time_probe, update_data=True)
            print("Emu_time not in cache: " + name)
            if cache:
                # Cached - make it read-only to prevent nasty overwriting bugs
                data.setflags(write=False)
                run_cache[emu_time_probe] = data

        if name not in run_cache:
            data = self.fetch_simdata(vcd_handle, name=name)
            print("Data not in cache: " + name)
            if cache:
                # Cached - make it read-only to prevent nasty overwriting bugs
                data.setflags(write=False)
                run_cache[name] = data
        else:
            data = run_cache[name]
            print("Data already in cache: " + name)

        if emu_time and name != emu_time_probe:
            print("Using emulation time")
            emutime_data = self.parse_emu_time(data=data, emu_time=run_cache[emu_time_probe])
            return emutime_data
        else:
            print("Using cycle counts as time basis")
            return data

    def _probes(self):
        """
        Get list of names probe waveforms for specified run

        :param run_num: Run number in sweep, omit/None for last run, 'all' for all runs
        :type run_num: int

        :rtype:list[str]
        """
        vcd_handle = self.setup_data_access()
        return vcd_handle.list_sigs()

    def discardloadedsimdatafiles(self):
        """

        """
        for key in self.vcd_handle.keys():
            vcd_handle = self.vcd_handle[key]
            try:
                del vcd_handle
                del self.probe_caches
            except:
                pass
        self.vcd_handle = dict()

    def setup_data_access(self):
        """
        :rtype: ParseVCD
        """
        if not self._data_valid:
            raise ValueError("No data available (no succesful simulation run / dataset reload)")

        #if simulator is None:
        #    simulator = self.simOpts._opts_container['simulator']

        #if simulator not in enums.SimulatorType.__dict__.keys():
        #    raise ValueError(
        #        "Specified Simulator is not valid. Valid simulators are: {0}; given simulator was: {1}".format(
        #            enums.SimulatorType, simulator))

        vcd_handle = "_".join([self.target._name])
        vcd_file_name = self.path_for_sim_result_file()
        if vcd_handle not in self.vcd_handle.keys():
            self.vcd_handle[vcd_handle] = ParseVCD(vcd_file_name)
        vcd_handle = self.vcd_handle[vcd_handle]

        return vcd_handle

    def path_for_sim_result_file(self):
        # Setup Simulation Result file names
        return os.path.join(self.target.cfg.vcd_path)

    def fetch_simdata(self, file_handle, name="", update_data=False):
        """
        Load VCD signals and store values as dictionary

        :return: Dict of numpy arrays (keys are column titles, values are column values)
        :rtype: dict[numpy.array]
        """
        signal = ''
        cycle_cnt = ''
        if name is not "":
            # parse only single signal name
            signal_dict = file_handle.parse_vcd(update_data=update_data)
            """ :type : dict()"""

            for key in signal_dict.keys():
                signal = [i[1] for i in signal_dict[key]['cv']]
                cycle_cnt = [i[0] for i in signal_dict[key]['cv']]

            if signal in [""]:
                raise ValueError("No data found for signal:{0}".format(name))
            return np.array([cycle_cnt, signal], dtype='O')

        else:
            # parse all signals into run_cache
            signal_dict = file_handle.parse_vcd(sigs=name, update_data=update_data)
            """ :type : dict()"""
            data = {}
            for key in signal_dict.keys():
                signal = [i[1] for i in signal_dict[key]['cv']]
                cycle_cnt = [i[0] for i in signal_dict[key]['cv']]
                net = signal_dict[key]['nets'][0]
                name = net['hier'] + '.' + net['name']
                data[name] = np.array([cycle_cnt,signal], dtype='O')
                data[name].setflags(write=False)

            if signal in [""]:
                raise ValueError("No data found for signal:{0}".format(name))

            return data




import os
import os.path
import json
import numpy as np

from argparse import ArgumentParser

from anasymod.config import EmuConfig, SimVisionConfig, XceliumConfig
from anasymod.sim.vivado import VivadoSimulator
from anasymod.sim.icarus import IcarusSimulator
from anasymod.sim.xcelium import XceliumSimulator
from anasymod.viewer.gtkwave import GtkWaveViewer
from anasymod.viewer.scansion import ScansionViewer
from anasymod.viewer.simvision import SimVisionViewer
from anasymod.emu.vivado_emu import VivadoEmulation
from anasymod.files import get_full_path, get_from_anasymod, mkdir_p
from anasymod.sources import VerilogSource
from anasymod.filesets import Filesets
from anasymod.defines import Define
from anasymod.targets import SimulationTarget, FPGATarget, Target
from anasymod.utils import statpro
from typing import Union
from importlib import import_module

class Analysis():
    """
    This is the top user Class that shall be used to exercise anasymod.
    """
    def __init__(self, input=None, build_root=None, simulator_name=None, synthesizer_name=None, viewer_name=None, preprocess_only=None, op_mode=None, active_target=None):

        # Indicate that setup is not finished yet
        self._setup_finished = False

        # Parse command line arguments
        self.args = None
        self._parse_args()

        # Overwrite input location in case it was provided when instantiation the Analysis class
        if input is not None:
            self.args.input = input

        # expand path of input and output directories relative to analysis.py
        self.args.input = get_full_path(self.args.input)

        # update args according to user specified values when instantiating analysis class
        self.args.simulator_name = simulator_name if simulator_name is not None else self.args.simulator_name
        self.args.synthesizer_name = synthesizer_name if synthesizer_name is not None else self.args.synthesizer_name
        self.args.viewer_name = viewer_name if viewer_name is not None else self.args.viewer_name
        self.args.preprocess_only = preprocess_only if preprocess_only is not None else self.args.preprocess_only
        self.args.active_target = active_target if active_target is not None else self.args.active_target

        self.default_fpga_target = 'fpga'
        self.default_sim_target = 'sim'

        if self.args.active_target == 'fpga':
            self.default_fpga_target = self.args.active_target
        elif self.args.active_target == 'sim':
            self.default_sim_target = self.args.active_target

        # Load config file
        try:
            self.cfg_file = json.load(open(os.path.join(self.args.input, 'prj_config.json'), 'r'))
        except:
            self.cfg_file = None
            print(f"Warning: no config file was fround for the project, expected path is: {os.path.join(self.args.input, 'prj_config.json')}")

        # Initialize project config
        self._prj_cfg = EmuConfig(root=self.args.input, cfg_file=self.cfg_file, build_root=build_root)

        # Initialize Plugins
        self._plugins = []
        for plugin in self._prj_cfg.cfg.plugins:
            try:
                i = import_module(f"plugin.{plugin}")
                inst = i.CustomPlugin(prj_cfg=self._prj_cfg, cfg_file=self.cfg_file, prj_root=self.args.input)
                self._plugins.append(inst)
                setattr(self, inst._name, inst)
            except:
                raise KeyError(f"Could not process plugin:{plugin} properly! Check spelling")

        # Check which mode is used to run, in case of commandline mode, besided setting up the class, also argument will be executed
        if op_mode in ['commandline']:
            print(f"Running in commandline mode.")

            # Finalize project setup, no more modifications of filesets and targets after that!!!
            self.setup_filesets()
            self.finish_setup()

            ###############################################################
            # Set options from to command line arguments
            ###############################################################

            self._prj_cfg.cfg.preprocess_only = self.args.preprocess_only

            ###############################################################
            # Execute actions according to command line arguments
            ###############################################################

            # generate bitstream
            if self.args.build:
                self.args.active_target = self.default_fpga_target if self.args.active_target is None else self.args.active_target
                self.build()

            # run FPGA if desired
            if self.args.emulate:
                self.args.active_target = self.default_fpga_target if self.args.active_target is None else self.args.active_target
                self.emulate()

            # launch FPGA if desired
            if self.args.launch:
                self.args.active_target = self.default_fpga_target if self.args.active_target is None else self.args.active_target
                self.launch()

            # run simulation if desired
            if self.args.sim or self.args.preprocess_only:
                self.args.active_target = self.default_sim_target if self.args.active_target is None else self.args.active_target
                self.simulate(unit=self.args.unit, id=self.args.id)

            # view results if desired
            if self.args.view and (self.args.sim or self.args.preprocess_only):
                self.args.active_target = self.default_sim_target if self.args.active_target is None else self.args.active_target
                self.view()

            if self.args.view and self.args.emulate:
                self.args.active_target = self.default_fpga_target if self.args.active_target is None else self.args.active_target
                self.view()

##### Functions exposed for user to exercise on Analysis Object

    def setup_filesets(self):
        """
        Setup filesets for project.
        This may differ from one project to another and needs customization.
        1. Read in source objects from source.config files and store those in a fileset object
        2. Add additional source objects to fileset object
        """

        # Read source.config files and store in fileset object
        default_filesets = ['default', 'sim', 'fpga']
        self.filesets = Filesets(root=self.args.input, default_filesets=default_filesets)
        self.filesets.read_filesets()

        # Add Defines and Sources from plugins
        for plugin in self._plugins:
            plugin._setup_sources()
            plugin._setup_defines()
            self.filesets._defines += plugin._dump_defines()
            self.filesets._verilog_sources += plugin._dump_verilog_sources()
            self.filesets._verilog_headers += plugin._dump_verilog_headers()
            self.filesets._vhdl_sources += plugin._dump_vhdl_sources()

        # Add custom source and define objects here e.g.:
        # self.filesets.add_source(source=VerilogSource())
        # self.filesets.add_define(define=Define())
        config_path = os.path.join(self.args.input, 'source.config')

        # Add some default files depending on whether there is a custom top level
        # TODO: clean this part up with generated top level
        self.filesets.add_source(source=VerilogSource(files=get_from_anasymod('verilog', 'vio_gen.sv'), config_path=config_path))
        for fileset in ['sim', 'fpga']:
            try:
                custom_top = self.cfg_file['TARGET'][fileset]['custom_top']
                print(f'Using custom top for fileset {fileset}.')
            except:
                custom_top = False

            if not custom_top:
                #ToDo: check if file inclusion should be target specific -> less for simulation only for example
                self.filesets.add_source(source=VerilogSource(files=os.path.join(self.args.input, 'tb.sv'), config_path=config_path, fileset=fileset))
                self.filesets.add_source(source=VerilogSource(files=os.path.join(self._prj_cfg.build_root, 'gen_vio_wrap.sv'), config_path=config_path, fileset=fileset))
                self.filesets.add_source(source=VerilogSource(files=os.path.join(self._prj_cfg.build_root, 'gen_ctrlwrap.sv'), config_path=config_path, fileset=fileset))
                self.filesets.add_source(source=VerilogSource(files=get_from_anasymod('verilog', 'gen_emu_clks.sv'), config_path=config_path, fileset=fileset))
                self.filesets.add_source(source=VerilogSource(files=get_from_anasymod('verilog', 'time_manager.sv'), config_path=config_path, fileset=fileset))

                get_from_anasymod('verilog', 'zynq_uart.bd')

        # Set define variables specifying the emulator control architecture
        # TODO: find a better place for these operations, and try to avoid directly accessing the config dictionary
        self.filesets.add_define(define=Define(name='DEC_BITS_MSDSL', value=self._prj_cfg.cfg.dec_bits))
        for fileset in ['sim', 'fpga']:
            try:
                top_module = self.cfg_file['TARGET'][fileset]['top_module']
            except:
                top_module = 'top'

            print(f'Using top module {top_module} for fileset {fileset}.')
            self.filesets.add_define(define=Define(name='CLK_MSDSL', value=f'{top_module}.emu_clk', fileset=fileset))
            self.filesets.add_define(define=Define(name='RST_MSDSL', value=f'{top_module}.emu_rst', fileset=fileset))
            self.filesets.add_define(define=Define(name='DEC_THR_MSDSL', value=f'{top_module}.emu_dec_thr', fileset=fileset))
            self.filesets.add_define(define=Define(name='DT_WIDTH', value=f'{self._prj_cfg.cfg.dt_width}', fileset=fileset))
            self.filesets.add_define(define=Define(name='DT_EXPONENT', value=f'{self._prj_cfg.cfg.dt_exponent}', fileset=fileset))
            self.filesets.add_define(define=Define(name='TIME_WIDTH', value=f'{self._prj_cfg.cfg.time_width}', fileset=fileset))

    def finish_setup(self):
        """
        Finalize filesets and setup targets. Both should not be modified anymore afterwards.
        :return:
        """

        # Initialize Targets
        self._setup_targets()

    def set_target(self, target_name):
        if self._setup_finished:
            self.args.active_target = target_name
            target = getattr(self, self.default_fpga_target)
            if isinstance(target, FPGATarget):
                self.default_fpga_target = target_name
            elif isinstance(target, SimulationTarget):
                self.default_sim_target = target_name
        else:
            raise Exception(f'Setup is not finished yet. Please finalize setup before setting the active target.')

    def build(self):
        """
        Generate bitstream for FPGA target
        """

        # Check if project setup was finished
        self._check_setup()

        # Check if active target is an FPGA target
        target = getattr(self, self.default_fpga_target)
        if not isinstance(target, FPGATarget):
            raise Exception(f'Active Target is of wrong type, only FPGATarget is supported for this action')

        VivadoEmulation(target=target).build()
        statpro.statpro_update(statpro.FEATURES.anasymod_build_vivado)

    def emulate(self, server_addr=None):
        """
        Program bitstream to FPGA and run simulation/emulation on FPGA
        """

        if server_addr is None:
            server_addr = self.args.server_addr

        # Check if project setup was finished
        self._check_setup()

        # Check if active target is an FPGA target
        target = getattr(self, self.default_fpga_target)
        if not isinstance(target, FPGATarget):
            raise Exception(f'Active Target is of wrong type, only FPGATarget is supported for this action')

        # create sim result folders
        if not os.path.exists(os.path.dirname(target.cfg.vcd_path)):
            mkdir_p(os.path.dirname(target.cfg.vcd_path))

        if not os.path.exists(os.path.dirname(target.cfg.csv_path)):
            mkdir_p(os.path.dirname(target.cfg.csv_path))

        # run the emulation
        VivadoEmulation(target=target).run_FPGA(start_time=self.args.start_time, stop_time=self.args.stop_time, server_addr=server_addr)
        statpro.statpro_update(statpro.FEATURES.anasymod_emulate_vivado)

        # post-process results
        from anasymod.wave import ConvertWaveform
        ConvertWaveform(target=target)

    def launch(self, server_addr=None):
        """
        Program bitstream to FPGA, setup control infrastructure and wait for interactive commands.
        :param server_addr: Address of Vivado hardware server used for communication to FPGA board
        :return:
        """

        if server_addr is None:
            server_addr = self.args.server_addr

        # Check if project setup was finished
        self._check_setup()

        # Check if active target is an FPGA target
        target = getattr(self, self.default_fpga_target)
        if not isinstance(target, FPGATarget):
            raise Exception(f'Active Target is of wrong type, only FPGATarget is supported for this action')

        # create sim result folders
        if not os.path.exists(os.path.dirname(target.cfg.vcd_path)):
            mkdir_p(os.path.dirname(target.cfg.vcd_path))

        if not os.path.exists(os.path.dirname(target.cfg.csv_path)):
            mkdir_p(os.path.dirname(target.cfg.csv_path))

        # launch the emulation
        ctrl_handle = VivadoEmulation(target=target).launch_FPGA(server_addr=server_addr)
        statpro.statpro_update(statpro.FEATURES.anasymod_emulate_vivado)

        # Return ctrl handle for interactive control
        return ctrl_handle

        #ToDo: once recording via ila in interactive mode is finishe and caotured results were dumped into a file,
        #ToDo: the conversion step to .vcd needs to be triggered via some command

    def simulate(self, unit=None, id=None):
        """
        Run simulation on a pc target.
        """

        # check if setup is already finished, if not do so
        if not self._setup_finished:
            self.finish_setup()

        # Check if active target is a SimulationTarget
        target = getattr(self, self.default_sim_target)
        if not isinstance(target, SimulationTarget):
            raise Exception(f'Active Target is of wrong type, only SimulationTarget is supported for this action')

        # create sim result folder
        if not os.path.exists(os.path.dirname(target.cfg.vcd_path)):
            mkdir_p(os.path.dirname(target.cfg.vcd_path))

        # pick simulator
        sim_cls = {
            'icarus': IcarusSimulator,
            'vivado': VivadoSimulator,
            'xrun': XceliumSimulator
        }[self.args.simulator_name]

        # run simulation

        sim = sim_cls(target=target)

        if self.args.simulator_name == "xrun":
            sim.unit = unit
            sim.id = id

        sim.simulate()
        statpro.statpro_update(statpro.FEATURES.anasymod_sim + self.args.simulator_name)

    def probe(self, name, emu_time=False):
        """
        Probe specified signal. Signal will be stored in a numpy array.
        """

        # Check if project setup was finished
        self._check_setup()

        target = getattr(self, self.args.active_target)

        probeobj = self._setup_probeobj(target=target)

        return probeobj._probe(name=name, emu_time=emu_time)

    def probes(self):
        """
        Display all signals that were stored for specified target run (simulation or emulation)
        :return: list of signal names
        """

        # Check if project setup was finished
        self._check_setup()

        target = getattr(self, self.args.active_target)

        probeobj = self._setup_probeobj(target=target)

        return probeobj._probes()

    def preserve(self, wave):
        """
        This function preserve the stepping of the waveform wave
        :param wave: 2d numpy.ndarray
        :return: 2d numpy.ndarray
        """
        temp_data = None
        wave_step =[]

        for d in wave.transpose():
            if temp_data is not None:
                if d[1] != temp_data:
                    wave_step.append([d[0],temp_data]) #old value with same timestep to preserve stepping
            wave_step.append(d)
            temp_data = d[1]

        try:
            return np.array(wave_step, dtype='float').transpose()
        except:
            return np.array(wave_step, dtype='O').transpose()

    def view(self):
        """
        View results from selected target run.
        """

        # Check if project setup was finished
        self._check_setup()

        target = getattr(self, self.args.active_target)

        # pick viewer
        viewer_cls = {
            'gtkwave': GtkWaveViewer,
            'simvision': SimVisionViewer,
            'scansion': ScansionViewer
        }[self.args.viewer_name]

        # set config file location for GTKWave
        # TODO: clean this up; it's a bit messy...
        if isinstance(target, FPGATarget):
            gtkw_search_order = ['view_fpga.gtkw', 'view.gtkw']
        elif isinstance(target, SimulationTarget):
            gtkw_search_order = ['view_sim.gtkw', 'view.gtkw']
        else:
            gtkw_search_order = ['view.gtkw']

        for basename in gtkw_search_order:
            candidate_path = os.path.join(self.args.input, basename)
            if os.path.isfile(candidate_path):
                self._prj_cfg.gtkwave_config.gtkw_config = candidate_path
                break
        else:
            self._prj_cfg.gtkwave_config.gtkw_config = None

        # set config file location for SimVision
        self._prj_cfg.simvision_config.svcf_config = os.path.join(self.args.input, 'view.svcf')

        # run viewer
        viewer = viewer_cls(target=target)
        viewer.view()

##### Utility Functions

    def _parse_args(self):
        """
        Read command line arguments. This supports convenient usage from command shell e.g.:
        python analysis.py -i filter --models --sim --view

        -i, --input: Path to project root directory of the project that shall be opened and worked with.
            default=None

        --simulator_name: Simulator that shall be used for logic simulation.
            default=icarus for windows, xrun for linux

        --synthesizer_name: Synthesis engine that shall be used for FPGA synthesis.
            default=vivado

        --viewer_name: Waveform viewer that shall be used for viewing result waveforms.
            default=gtkwave for windows, simvision for linux

        --active_target: Target that shall be actively used.
            default='sim'

        --launch: Launch the FPGA simulation/emulation by programming the bitstream and preparing the control interface for interactive use.

        --sim: Execute logic simulation for selected simulation target.

        --view: Open results in selected waveform viewer.

        --build: Synthesize, run P&R and generate bitstream for selected target.

        --emulate: Execute FPGA run for selected target.

        --start_time: Start time for FPGA simulation.
            default=0

        --server_addr: Hardware server address for FPGA simulation. This is necessary for connecting to a vivado
            hardware server from linux, that was setup under windows.
            default=None

        --stop_time: Stop time for FPGA simulation
            default=None

        --preprocess_only: For icarus only, this will nur run the simulation, but only compile the netlist.

        """

        parser = ArgumentParser()

        # if the Cadence tools are available, use those as defaults instead
        try:
            x = XceliumConfig(None).xrun
            default_simulator_name = 'xrun' if x is not None else 'icarus'
        except:
            default_simulator_name = 'icarus'

        try:
            s = SimVisionConfig(None).simvision
            default_viewer_name = 'simvision' if s is not None else 'gtkwave'
        except:
            default_viewer_name = 'gtkwave'
            pass


        parser.add_argument('-i', '--input', type=str, default=None)
        parser.add_argument('--simulator_name', type=str, default=default_simulator_name)
        parser.add_argument('--synthesizer_name', type=str, default='vivado')
        parser.add_argument('--viewer_name', type=str, default=default_viewer_name)
        parser.add_argument('--active_target', type=str, default=None)
        parser.add_argument('--unit', type=str, default=None)
        parser.add_argument('--id', type=str, default=None)
        parser.add_argument('--sim', action='store_true')
        parser.add_argument('--view', action='store_true')
        parser.add_argument('--build', action='store_true')
        parser.add_argument('--emulate', action='store_true')
        parser.add_argument('--launch', action='store_true')
        parser.add_argument('--start_time', type=float, default=0)
        parser.add_argument('--server_addr', type=str, default=None)
        parser.add_argument('--stop_time', type=float, default=None)
        parser.add_argument('--preprocess_only', action='store_true')

        self.args, _ = parser.parse_known_args()

    def _setup_targets(self):
        """
        Setup targets for project.
        This may differ from one project to another and needs customization.
        1. Create target object for each target that is supported in project
        2. Assign filesets to all target objects of the project
        """

        # Populate the fileset dict which will be used to copy data to target object and store in filesets variable
        self.filesets.populate_fileset_dict()
        filesets = self.filesets.fileset_dict

        #######################################################
        # Create and setup simulation target
        #######################################################

        self.sim = SimulationTarget(prj_cfg=self._prj_cfg, plugins=self._plugins, name=r"sim")
        self.sim.assign_fileset(fileset=filesets['default'])
        if 'sim' in filesets:
            self.sim.assign_fileset(fileset=filesets['sim'])



        # Update simulation target specific configuration
        self.sim.cfg.update_config(subsection=r"sim")
        self.sim.update_structure_config()
        #ToDo: currently structure generation is only necessary for FPGA simulation, in future at least signals selected in probing file should automatically be included in vcd file
        if not self.sim.cfg.custom_top:
            print('!!! gen_structure for simulation target')
            self.sim.setup_ctrl_ifc()
            self.sim.gen_structure()
        self.sim.set_tstop()
        self.sim.setup_vcd()

        #######################################################
        # Create and setup FPGA target
        #######################################################
        self.fpga = FPGATarget(prj_cfg=self._prj_cfg, plugins=self._plugins, name=r"fpga")
        self.fpga.assign_fileset(fileset=filesets['default'])
        if 'fpga' in filesets:
            self.fpga.assign_fileset(fileset=filesets['fpga'])

        # Update simulation target specific configuration
        self.fpga.cfg.update_config(subsection=r"fpga")
        self.fpga.update_structure_config()
        # Instantiate Simulation ControlInfrastructure Interface
        # ToDo: This should only be executed for FPGA simulation!
        if not self.fpga.cfg.custom_top:
            print('!!! gen_structure for FPGA target')
            self.fpga.setup_ctrl_ifc()
            self.fpga.gen_structure()
        self.fpga.set_tstop()

        # Set indication that project setup is complete
        self._setup_finished = True

    def _setup_probeobj(self, target: Union[FPGATarget, SimulationTarget]):
        """
        Check if the requested probe obj in the target object already exists, in not create one.
        Return the probe object.

        :param target: Target that signals shall be extracted from
        :return: probe object that was selected in target object
        """

        # specify probe obj name, specific to selected simulator/synthesizer
        if isinstance(target, FPGATarget):
            target_name = f"prb_{self.args.synthesizer_name}"
        elif isinstance(target, SimulationTarget):
            target_name = f"prb_{self.args.simulator_name}"
        else:
            raise ValueError(f"Provided target type:{target} is not supported")

        # check if probe obj is already existing, if not, instantiate one

        #ToDo: In future it should be also possible to instantiate different probe objects, depending on data format that shall be read in
        if target_name not in target.probes.keys():
            from anasymod.probe import ProbeVCD
            target.probes[target_name] = ProbeVCD(target=target)

        return target.probes[target_name]

    def _check_setup(self):
        """
        Check if the targets were already created for the project. If not, throw an error.

        :return:
        """

        if not self._setup_finished:
            raise ValueError("The project setup changed after data aquisition; Data might be out of sync!")

def main():
    Analysis(op_mode='commandline')

if __name__ == '__main__':
    main()



import os

from glob import glob
from anasymod.generators.codegen import CodeGenerator
from anasymod.util import back2fwd
from typing import Union

class ConfigFileObj(CodeGenerator):
    def __init__(self, files, config_path):
        super().__init__()
        if isinstance(files, list):
            self.files = files
        elif isinstance(files, str):
            self.files = [files]
        else:
            raise TypeError(f"Type of config_paths variable provided to SubConfig class is not a list, is:{type(files)} instead.")

        self.config_path = config_path

        # In case config_path was not specified yet (default is None) path expansion needs to be performed later manually after setting config_path
        # This is necessary to improve conveniency of providing sources in config file
        if config_path is not None:
            self.expand_paths()

    def expand_paths(self):
        """
        Expand environment variables in provided list of paths.
        Check if path is absolute or relative, in case of a relative path, it will be expanded to an absolute path,
        whereas the folder of the config_file will be used to complete the path.
        """
        abs_paths = []
        if not isinstance(self.files, list):
            raise TypeError(f"Wrong format used in config file {self.config_path}")
        for file in self.files:
            file = os.path.expandvars(file.strip('" '))
            if not os.path.isabs(file):
                if self.config_path is not None:
                    path_suffix = file.replace('\\', '/').replace('/', os.sep).split(os.sep)
                    try:
                        path_suffix.remove('.')
                    except:
                        pass
                    file = os.path.join(os.path.dirname(self.config_path), *(path_suffix))
                else:
                    raise KeyError(f"No config path was provided, is set to:{self.config_path}")
            abs_paths.append(file)

        self.files = [file for p in abs_paths for file in glob(p)]

class SubConfig(ConfigFileObj):
    def __init__(self, files: Union[list, str], config_path=None):
        super().__init__(files=files, config_path=config_path)

class Sources(ConfigFileObj):
    def __init__(self, files: list, fileset, config_path):
        super().__init__(files=files, config_path=config_path)
        self.fileset = fileset

    def generate(self):
        pass

    def set_property(self, name, value, objects):
        self.writeln(' '.join(['set_property', '-name', name, '-value', value, '-objects', objects]))

class VerilogSource(Sources):
    def __init__(self, files: Union[list, str], fileset=r"default", config_path=None, verilog_version=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path)
        self.verilog_version = verilog_version

    def generate(self):
        self.text = self.files
        return self.dump()

class VerilogHeader(Sources):
    def __init__(self, files: Union[list, str], fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path)

    def set_header_files(self):
        file_list = '{ ' + ' '.join('"' + back2fwd(file) + '"' for file in self.files) + ' }'
        self.set_property('file_type', '{Verilog Header}', f'[get_files {file_list}]')

    def generate(self):
        self.dump()

class VHDLSource(Sources):
    def __init__(self, files: Union[list, str], library=None, fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path)
        self.library = library

    def set_vhdl_library(self):
        file_list = '{ ' + ' '.join('"' + back2fwd(file) + '"' for file in self.files) + ' }'
        self.set_property('library', value=self.library, objects=f'[get_files {file_list}]')

    def generate(self):
        self.dump()

class XCIFile(Sources):
    def __init__(self, files: Union[list, str], fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path)

class XDCFile(Sources):
    def __init__(self, files: Union[list, str], fileset=r"default", config_path=None):
        super().__init__(files=files, fileset=fileset, config_path=config_path)

class MEMFile(Sources):
    def __init__(self, files: str, fileset=r"default", config_path=None):
        super().__init__(files=[files], fileset=fileset, config_path=config_path)

class BDFile(Sources):
    def __init__(self, files: str, fileset=r"default", config_path=None):
        super().__init__(files=[files], fileset=fileset, config_path=config_path)


from setuptools import setup

name = 'anasymod'
version = '0.1.3'

DESCRIPTION = '''\
Tool for running mixed-signal emulations on FPGAs\
'''

with open('README.md', 'r') as fh:
    LONG_DESCRIPTION = fh.read()

setup(
    name=name,
    version=version,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords = ['analog', 'mixed-signal', 'mixed signal', 'vivado',
                'xcelium', 'iverilog', 'icarus-verilog', 'icarus verilog',
                'generator', 'verilog', 'system-verilog', 'system verilog',
                'emulation', 'fpga'],
    packages=[
        f'{name}'
    ],
    entry_points = {
        'console_scripts': [
            'anasymod=anasymod.analysis:main'
        ]
    },
    install_requires=[
        'svreal',
        'msdsl',
        'jinja2',
        'pyvcd',
        'pyserial'
    ],
    license='BSD 3-Clause "New" or "Revised" License',
    url=f'https://github.com/sgherbst/{name}',
    author='Gabriel Rutsch, Steven Herbst, Shivani Saravanan',
    author_email='gabriel.rutsch@infineon.com',
    python_requires='>=3.7',
    download_url = f'https://github.com/sgherbst/{name}/archive/v{version}.tar.gz',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'License :: OSI Approved :: BSD License',
        f'Programming Language :: Python :: 3.7'
    ],
    include_package_data=True,
    zip_safe=False
)



import os
import os.path
import subprocess
import sys
import json

from multiprocessing.pool import ThreadPool
from math import ceil, log2
from collections import namedtuple
from argparse import ArgumentParser

from msdsl import VerilogGenerator

def back2fwd(path: str):
    return path.replace('\\', '/')

def call(args, cwd=None, wait=True):
    # set defaults
    if cwd is None:
        cwd = os.getcwd()

    # remember starting directory and change to cwd
    start_dir = os.getcwd()
    os.chdir(cwd)

    # run the command
    kwargs = dict(args=args, stdout=sys.stdout, stderr=sys.stdout)
    if wait:
        ret = subprocess.call(**kwargs)
        assert ret == 0
    else:
        subprocess.Popen(**kwargs)

    # change back to the starting directory
    os.chdir(start_dir)

def next_pow_2(x):
    return 2**int(ceil(log2(x)))

########################
# parallel_scripts
# ref: https://stackoverflow.com/questions/26774781/python-multiple-subprocess-with-a-pool-queue-recover-output-as-soon-as-one-finis

def parallel_calls(calls, num=None):
    tp = ThreadPool(num)

    for arg_list in calls:
        tp.apply_async(call, (arg_list,))

    tp.close()
    tp.join()

########################

########################
# file_len: modified from https://stackoverflow.com/questions/845058/how-to-get-line-count-cheaply-in-python

def file_len(fname):
    with open(fname, encoding='utf-8') as f:
        i = 0

        for i, l in enumerate(f, 1):
            pass

        return i
########################

def vivado_search_key(dir_):
    year, version = os.path.basename(dir_).split('.')
    year, version = int(year), int(version)

    return -year, -version

########################
# parallel_scripts
# ref: https://stackoverflow.com/questions/26774781/python-multiple-subprocess-with-a-pool-queue-recover-output-as-soon-as-one-finis

def parallel_calls(calls, num=None):
    tp = ThreadPool(num)

    for arg_list in calls:
        tp.apply_async(call, (arg_list,))

    tp.close()
    tp.join()

########################

########################
# file_len: modified from https://stackoverflow.com/questions/845058/how-to-get-line-count-cheaply-in-python

def file_len(fname):
    with open(fname, encoding='utf-8') as f:
        i = 0

        for i, l in enumerate(f, 1):
            pass

        return i
########################

########################
# JSON to object: from https://stackoverflow.com/questions/6578986/how-to-convert-json-data-into-a-python-object

def _json_object_hook(d):
    return namedtuple('X', d.keys())(*d.values())
def json2obj(data):
    return json.loads(data, object_hook=_json_object_hook)
########################

# Argument parser for the examples
class ExampleControl:
    def __init__(self):
        # create the parser
        parser = ArgumentParser()

        # add custom arguments
        parser.add_argument('-o', '--output', type=str)
        parser.add_argument('--dt', type=float)

        # parser arguments
        args = parser.parse_args()

        # save arguments
        self.output = os.path.realpath(os.path.expanduser(args.output))
        self.dt = args.dt

    def write_model(self, model):
        # determine the filename
        filename = os.path.join(self.output, f'{model.module_name}.sv')
        print('Model will be written to: ' + filename)

        # write the model to a file
        model.compile_to_file(VerilogGenerator(), filename)

def main():
    print(next_pow_2(15))
    print(next_pow_2(16))
    print(next_pow_2(17))

if __name__ == '__main__':
    main()



import os
from anasymod.sources import Sources, VerilogSource, VerilogHeader, VHDLSource, SubConfig, XCIFile, XDCFile, MEMFile, BDFile
from anasymod.defines import Define

class Filesets():
    def __init__(self, root, default_filesets=['default', 'sim', 'fpga']):
        self._master_cfg_path = os.path.join(root, 'source.config')

        self._verilog_sources = []
        """:type : List[VerilogSource]"""

        self._verilog_headers = []
        """:type : List[VerilogHeader]"""

        self._vhdl_sources = []
        """:type : List[VHDLSource]"""

        self._xci_files = []
        """:type : List[XCIFile]"""

        self._xdc_files = []
        """:type : List[XDCFile]"""

        self._defines = []
        """:type : List[Define]"""

        self._mem_files = []
        """:type : List[MEMFile]"""

        self._bd_files = []
        """:type : List[BDFile]"""

        # init fileset_dict
        self.fileset_dict = {}
        if not default_filesets:
            for fileset in default_filesets:
                self.fileset_dict[fileset] = {}

        self._sub_config_paths = []
        """:type : List[SubConfig]"""

    def read_filesets(self, validate_paths=False):
        if os.path.isfile(self._master_cfg_path):
            with open(self._master_cfg_path, "r") as f:
                mcfg = f.readlines()

            # Read source paths from master config
            self._parseconfig(cfg=mcfg, cfg_path=self._master_cfg_path)

            # Read source paths from sub configs
            while (bool(self._sub_config_paths)):
                for config in self._sub_config_paths:
                    self._sub_config_paths.remove(config)
                    for file in config.files:
                        if os.path.isfile(file):
                            with open(file, "r") as f:
                                cfg = f.readlines()
                            self._parseconfig(cfg=cfg, cfg_path=file)
                        else:
                            print(f"WARNING: provided path:'{config}' does not exist, skipping config file")

            # Check if fileset paths exist
            if validate_paths:
                for key in self.fileset_dict.keys():
                    for path in self.fileset_dict[key]:
                        if not os.path.exists(path):
                            raise ValueError(f"Provided path:{path} of fileset:{key} does not exist")
        else:
            print(f"No config file existing, skipping to read source files.")

    def _parseconfig(self, cfg: list, cfg_path: str):
        """
        Read all lines from config file, according to string infront of '=' sign, the proceeding arguments will either
        be added to the according fileset or will be added to the list of additional config paths, which will be
        investigated in the next iteration.
        :param cfg:
        :param cfg_path:
        :return:
        """

        for k, line in enumerate(cfg):
            line = line.strip()

            if line.startswith('#'):
                # skip comments
                continue

            if line:
                try:
                    line = eval(line)
                    if isinstance(line, VerilogSource):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._verilog_sources.append(line)
                    elif isinstance(line, VerilogHeader):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._verilog_headers.append(line)
                    elif isinstance(line, VHDLSource):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._vhdl_sources.append(line)
                    elif isinstance(line, XCIFile):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._xci_files.append(line)
                    elif isinstance(line, XDCFile):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._xdc_files.append(line)
                    elif isinstance(line, MEMFile):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._mem_files.append(line)
                    elif isinstance(line, BDFile):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._bd_files.append(line)
                    elif isinstance(line, SubConfig):
                        line.config_path = cfg_path
                        line.expand_paths()
                        self._sub_config_paths.append(line)
                    elif isinstance(line, Define):
                        self._defines.append(line)
                    else:
                        raise Exception(f"Line {k+1} of config file: {cfg_path} does not fit do a specified source or config type")
                except:
                    raise Exception(f"Line {k+1} of config file: {cfg_path} could not be processed properely")

    def populate_fileset_dict(self):
        """
        Creates fileset dictionary according to filesets that were provided reading in source and define objects.
        Previously created filesets will be overwritten.
        """

        # Read in verilog source objects to fileset dict
        self._add_to_fileset_dict(name='verilog_sources', container=self._verilog_sources)

        # Read in verilog header objects to fileset dict
        self._add_to_fileset_dict(name='verilog_headers', container=self._verilog_headers)

        # Read in vhdlsource objects to fileset dict
        self._add_to_fileset_dict(name='vhdl_sources', container=self._vhdl_sources)

        # Read in xcifile objects to fileset dict
        self._add_to_fileset_dict(name='xci_files', container=self._xci_files)

        # Read in xdcfile objects to fileset dict
        self._add_to_fileset_dict(name='xdc_files', container=self._xdc_files)

        # Read in define objects to fileset dict
        self._add_to_fileset_dict(name='defines', container=self._defines)

        # Read in mem_files objects to fileset dict
        self._add_to_fileset_dict(name='mem_files', container=self._mem_files)

        # Read in bd_files objects to fileset dict
        self._add_to_fileset_dict(name='bd_files', container=self._bd_files)

    def _add_to_fileset_dict(self, name, container):
        """
        Adds a specified attribute to the fileset_dict, e.g. add the verilog sources or defines.
        """
        for item in container:
            if item.fileset in self.fileset_dict.keys():
                if name in self.fileset_dict[item.fileset]:
                    self.fileset_dict[item.fileset][name].append(item)
                else:
                    self.fileset_dict[item.fileset][name] = [item]
            else:
                print(f"Custom fileset was added:{item.fileset}")
                self.fileset_dict[item.fileset] = {}
                self.fileset_dict[item.fileset][name] = [item]


    def add_source(self, source: Sources):
        if isinstance(source, VerilogSource):
            self._verilog_sources.append(source)
        if isinstance(source, VerilogHeader):
            self._verilog_headers.append(source)
        if isinstance(source, VHDLSource):
            self._vhdl_sources.append(source)

    def add_define(self, define: Define):
        self._defines.append(define)

    def add_xci_file(self, xci_file: XCIFile):
        self._xci_files.append(xci_file)

    def add_xdc_file(self, xdc_file: XDCFile):
        self._xdc_files.append(xdc_file)

    def add_mem_file(self, mem_file: MEMFile):
        self._mem_files.append(mem_file)

    def add_bd_file(self, bd_file: BDFile):
        self._bd_files.append(bd_file)

def main():
    fileset = Filesets(root=r"C:\Inicio_dev\anasymod\tests\filter")
    fileset.read_filesets()
    print(fileset.fileset_dict)

if __name__ == '__main__':
    main()